import logging
import pandas as pd
import numpy as np
from scipy.signal import find_peaks, peak_widths
from fraggler.ladder_fitting.fit_ladder_model import FitLadderModel


######### -------------------- Custom Errors ######### --------------------
class OverlappingIntervalError(Exception):
    pass


class WrongColumnsError(Exception):
    pass


######### -------------------- Validation functions ######### --------------------
def is_overlapping(df: pd.DataFrame) -> bool:
    test = (
        df.sort_values("start")
        .assign(intervals=lambda x: [range(y.start, y.stop) for y in x.itertuples()])
        .explode("intervals")
    )

    if test.shape[0] != test.intervals.nunique():
        dups = (
            test.value_counts("intervals")
            .reset_index()
            .sort_values("intervals")
            .loc[lambda x: x["count"] > 1]
            .iloc[0, 0]
        )
        logging.warning(
            f"""
        Customized peaks contains overlapping ranges
        Starting at value: {dups}
        """
        )
        return True
    return False


def has_columns(df: pd.DataFrame) -> bool:
    columns = set(
        ["name", "start", "stop", "amount", "min_ratio", "which", "peak_distance"]
    )
    df_columns = set(df.columns)

    if len(columns) != len(df_columns):
        logging.warning(
            f"""
        Customized peaks table does not contain the right columns.
        Current columns: {df_columns}
        Needed columns: {columns}
        """
        )
        return False

    intersection = columns.intersection(df_columns)
    if len(intersection) != len(df_columns):
        logging.warning(
            f"""
        Customized peaks table does not contain the right columns.
        Current columns: {df_columns}
        Needed columns: {columns}
        """
        )
        return False

    return True


#### ---------------------------------------------------------------------------------- ###


class PeakFinder:
    def __init__(
        self,
        model: FitLadderModel,
        min_ratio: float = 0.15,
        search_peaks_start: int = 110,
        peak_height: int = 350,
        distance_between_assays: int = 15,
        custom_peaks: str | pd.DataFrame = None,
    ) -> None:
        self.model = model
        self.raw_data = self.model.adjusted_baisepair_df
        self.file_name = self.model.fsa_file.file_name
        self.search_peaks_start = search_peaks_start
        self.peak_height = peak_height
        self.min_ratio = min_ratio
        self.distance_between_assays = distance_between_assays
        self.custom_peaks = (
            custom_peaks.fillna(0)
            if isinstance(custom_peaks, pd.DataFrame)
            else pd.read_csv(custom_peaks).fillna(0)
            if isinstance(custom_peaks, str)
            else None
        )

        # Validation of custom_peaks
        self.run_validation()
        # find peaks, custom or agnostic
        self.find_peaks()
        # continue whether peaks are found or not.
        self.found_peaks()

    def run_validation(self):
        if self.custom_peaks is None:
            return

        if is_overlapping(self.custom_peaks):
            raise OverlappingIntervalError("Overlapping intervals!")
            exit(1)

        if not has_columns(self.custom_peaks):
            raise WrongColumnsError("Wrong columns!")
            exit(1)

    def find_peaks(self):
        if self.custom_peaks is not None:
            self.find_peaks_customized(
                peak_height=self.peak_height,
            )
        else:
            self.find_peaks_agnostic(
                peak_height=self.peak_height,
                min_ratio=self.min_ratio,
                distance_between_assays=self.distance_between_assays,
            )

    def found_peaks(self):
        # if no peaks could be found
        if self.peak_information.shape[0] == 0:
            self.found_peaks = False
            logging.warning(f"No peaks could be found. Please look at raw data.")
        # if peaks are found
        else:
            self.found_peaks = True

    def find_peaks_agnostic(
        self,
        peak_height: int,
        min_ratio: float,
        distance_between_assays: int,
    ) -> None:
        peaks_dataframe = self.raw_data.loc[
            lambda x: x.basepairs > self.search_peaks_start
        ]
        peaks_index, _ = find_peaks(peaks_dataframe.peaks, height=peak_height)

        peak_information = (
            peaks_dataframe.iloc[peaks_index]
            .assign(peaks_index=peaks_index)
            .assign(peak_name=lambda x: range(1, x.shape[0] + 1))
            # separate the peaks into different assay groups depending on the distance
            # between the peaks
            .assign(difference=lambda x: x.basepairs.diff())
            .fillna(100)
            .assign(
                assay=lambda x: np.select(
                    [x.difference > distance_between_assays],
                    [x.peak_name * 10],
                    default=pd.NA,
                )
            )
            .fillna(method="ffill")
            .assign(max_peak=lambda x: x.groupby("assay")["peaks"].transform(np.max))
            .assign(ratio=lambda x: x.peaks / x.max_peak)
            .loc[lambda x: x.ratio > min_ratio]
            .assign(peak_name=lambda x: range(1, x.shape[0] + 1))
        )

        # update peaks_index based on the above filtering
        peaks_index = peak_information.peaks_index.to_numpy()

        # update class attributes
        self.peaks_index = peaks_index
        self.peaks_dataframe = peaks_dataframe
        self.peak_information = peak_information

    def find_peaks_customized(
        self,
        peak_height: int,
    ) -> None:

        # Filter where to start search for the peaks
        peaks_dataframe = self.raw_data.loc[
            lambda x: x.basepairs > self.search_peaks_start
        ]
        # Find the peaks
        peaks_index, _ = find_peaks(peaks_dataframe.peaks, height=peak_height)

        # Filter the df to get right peaks
        peak_information = peaks_dataframe.iloc[peaks_index].assign(
            peaks_index=peaks_index
        )
        # Filter the above df based on the custom peaks from the user
        customized_peaks = []
        for assay in self.custom_peaks.itertuples():
            df = (
                peak_information.loc[lambda x: x.basepairs > assay.start]
                .loc[lambda x: x.basepairs < assay.stop]
                .assign(assay=assay.name)
            )

            # Rank the peaks by height and filter out the smallest ones
            if assay.amount != 0:
                if assay.which == "LARGEST" or assay.which == "":
                    df = (
                        df.assign(max_peak=lambda x: x.peaks.max())
                        .assign(ratio=lambda x: x.peaks / x.max_peak)
                        .loc[lambda x: x.ratio > assay.min_ratio]
                        .assign(rank_peak=lambda x: x.peaks.rank(ascending=False))
                        .loc[lambda x: x.rank_peak <= assay.amount]
                        .drop(columns=["rank_peak"])
                    )
                    if assay.peak_distance != 0:
                        df = (
                            df.assign(distance=lambda x: x.basepairs.diff())
                            .assign(distance=lambda x: x.distance.fillna(0))
                            .loc[lambda x: x.distance <= assay.peak_distance]
                            .drop(columns=["distance"])
                        )

                elif assay.which == "FIRST":
                    df = (
                        df.assign(max_peak=lambda x: x.peaks.max())
                        .assign(ratio=lambda x: x.peaks / x.max_peak)
                        .loc[lambda x: x.ratio > assay.min_ratio]
                        .sort_values("basepairs", ascending=True)
                        .head(assay.amount)
                    )
                    if assay.peak_distance != 0:
                        df = (
                            df.assign(distance=lambda x: x.basepairs.diff())
                            .assign(distance=lambda x: x.distance.fillna(0))
                            .loc[lambda x: x.distance <= assay.peak_distance]
                            .drop(columns=["distance"])
                        )
                else:
                    print("[ERROR]: column `which` must be `FIRST` or `LARGEST`")
                    exit(1)

            customized_peaks.append(df)

        peak_information = (
            pd.concat(customized_peaks)
            .reset_index()
            .assign(peak_name=lambda x: range(1, x.shape[0] + 1))
        )

        # update peaks_index based on the above filtering
        peaks_index = peak_information.peaks_index.to_numpy()

        # update class attributes
        self.peaks_index = peaks_index
        self.peaks_dataframe = peaks_dataframe
        self.peak_information = peak_information



r"""
.. include:: ../README.md
.. include:: ../CONTRIBUTION.md
"""

__author__ = "William Rosenbaum and Pär Larsson"

from .ladder_fitting.fit_ladder_model import FitLadderModel
from .ladder_fitting.peak_ladder_assigner import PeakLadderAssigner
from .ladders.ladders import LADDERS
from .plotting.plot_ladder import PlotLadder
from .utils.baseline_removal import baseline_arPLS
from .utils.fsa_file import FsaFile
from .utils.utils import get_files, setup_logging
from .utils.fraggler_object import (
    FragglerPeak,
    FragglerArea,
    make_fraggler_peak,
    make_fraggler_area,
)
from .applications.peak_area_multiplex import PeakAreaDeMultiplex
from .plotting.plot_peak_area import PlotPeakArea
from .plotting.plot_raw_data import PlotRawData
from .plotting.plot_peaks import PlotPeaks
from .plotting.plot_channels import make_fsa_data_df, plot_fsa_data
from .reports.reports import (
    generate_peak_report,
    generate_area_report,
    generate_no_peaks_report,
)
from .functions.generate_peak_table import generate_peak_table
from .utils.peak_finder import PeakFinder
from .cli import peak_report, area_report


__all__ = [
    "FitLadderModel",
    "PeakLadderAssigner",
    "LADDERS",
    "baseline_arPLS",
    "FsaFile",
    "PeakArea",
    "PeakAreaDeMultiplex",
    "PlotPeakArea",
    "PlotLadder",
    "PlotRawData",
    "PlotPeaks",
    "generate_peak_table",
    "get_files",
    "setup_logging",
    "make_fsa_data_df",
    "plot_fsa_data",
    "PeakFinder",
    "FragglerPeak",
    "FragglerArea",
    "make_fraggler_peak",
    "make_fraggler_area",
    "generate_peak_report",
    "generate_area_report",
    "generate_no_peaks_report",
    "peak_report",
    "area_report",
]



import pandas as pd
import altair as alt


def make_fsa_data_df(fsa, ladder: bool = False) -> pd.DataFrame:
    data = ["DATA1", "DATA2", "DATA3", "DATA4"]
    if ladder:
        data.append("DATA205")
    dfs = []
    for d in data:
        df = (
            pd.DataFrame()
            .assign(data=fsa.fsa[d])
            .assign(channel=d)
            .assign(time=lambda x: range(x.shape[0]))
        )
        dfs.append(df)
    return pd.concat(dfs)


def plot_fsa_data(fsa: str, ladder: bool = False) -> list:
    alt.data_transformers.disable_max_rows()
    df = make_fsa_data_df(fsa, ladder)

    plots = []
    for channel in df.channel.unique():
        plot = (
            alt.Chart(df.loc[lambda x: x.channel == channel])
            .mark_line()
            .encode(
                alt.X("time:Q", title="Time"),
                alt.Y("data:Q", title="Intensity"),
                alt.Color("channel:N"),
            )
            .properties(
                width=800,
                height=500,
            )  # .interactive()
        )
        plots.append((plot, channel))

    all_data = (
        alt.Chart(df)
        .mark_line()
        .encode(
            alt.X("time:Q", title="Time"),
            alt.Y("data:Q", title="Intensity"),
            alt.Color("channel:N"),
        )
        .properties(
            width=800,
            height=500,
        )  # .interactive()
    )
    plots.append((all_data, "All channels"))
    return plots



import pandas as pd
from pathlib import Path
import fraggler


# TODO just take Fraggler object
def generate_peak_table(
    in_files: str | list,
) -> pd.DataFrame:
    if isinstance(in_files, str):
        in_files = [x for x in Path(in_files).iterdir() if x.suffix == ".fsa"]

    peak_dfs = []
    for x in in_files:
        try:
            fsa = fraggler.FsaFile(
                x, ladder, min_height=min_height, trace_channel=trace_channel
            )
            pla = fraggler.PeakLadderAssigner(fsa)
            model = fraggler.FitLadderModel(pla)
            pam = fraggler.PeakAreaDeMultiplex(
                model,
                cutoff=cutoff,
                min_ratio=min_ratio,
                custom_peaks=custom_peaks,
            )
            peak_dfs.append(pam.assays_dataframe(peak_model))
        except:
            print(f"FAILED: {fsa.file_name}")

    return pd.concat(peak_dfs).reset_index(drop=True)



import logging
import re
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from lmfit.models import VoigtModel, GaussianModel, LorentzianModel
from scipy.signal import find_peaks, peak_widths
from ..utils.peak_finder import PeakFinder


class PeakAreaDeMultiplexIterator:
    def __init__(self, number_of_assays):
        self.number_of_assays = number_of_assays
        self.current = 0

    def __next__(self):
        if self.current >= self.number_of_assays:
            raise StopIteration
        result = self.current
        self.current += 1
        return result


class PeakAreaDeMultiplex:
    def __init__(
        self,
        peaks: PeakFinder,
        cutoff: float = None,
    ) -> None:
        self.peaks = peaks
        self.cutoff = cutoff or None
        self.peaks_dataframe = self.peaks.peaks_dataframe
        self.peak_information = self.peaks.peak_information
        self.file_name = self.peaks.file_name
        self.peaks_index = self.peaks.peaks_index
        self.found_peaks = self.peaks.found_peaks
        self.area_plots = []

        self.find_peak_widths()
        # divade peaks based on their assay they belonging
        self.divided_assays = self.divide_peak_assays()
        # how many assays does this sample contain?
        self.number_of_assays = len(self.divided_assays)
        # divide all peaks in each assay into separate dataframes
        self.divided_peaks = [self.divide_peaks(x) for x in self.divided_assays]
        # logging
        logging.info(
            f"""
        Number of assays found: {self.number_of_assays}
        Number of peaks found: {self.peak_information.shape[0]}
        """
        )

    def __iter__(self):
        return PeakAreaDeMultiplexIterator(self.number_of_assays)

    def find_peak_widths(self, rel_height: float = 0.95):
        widths = peak_widths(
            self.peaks_dataframe.peaks,
            self.peaks_index,
            rel_height=rel_height,
        )

        df = pd.DataFrame(widths).T
        df.columns = ["x", "peak_height", "peak_start", "peak_end"]

        self.peak_widths = (
            df.assign(peak_start=lambda x: np.floor(x.peak_start).astype(int))
            .assign(peak_end=lambda x: np.ceil(x.peak_end).astype(int))
            .assign(peak_name=lambda x: range(1, x.shape[0] + 1))
            .merge(self.peak_information, on="peak_name")
        )

    def divide_peak_assays(self) -> list[pd.DataFrame]:
        """
        Divide the peaks belonging to different assays based on their assay number
        """
        df = self.peak_widths
        return [df.loc[df.assay == x] for x in df.assay.unique()]

    def divide_peaks(self, assay: pd.DataFrame, padding: int = 4) -> list[pd.DataFrame]:
        # add some padding to the left and right to be sure to include everything in the peak
        return [
            self.peaks_dataframe.iloc[x.peak_start - padding : x.peak_end + padding]
            for x in assay.itertuples()
        ]

    def fit_lmfit_model(self, peak_finding_model: str, assay_number: int):
        if assay_number >= self.number_of_assays:
            raise IndexError(
                f"""
                The sample only contains {self.number_of_assays} assays. 
                Use a number inside of the range.
                Indexing starts at 0.
                """
            )

        if peak_finding_model == "gauss":
            model = GaussianModel()
        elif peak_finding_model == "voigt":
            model = VoigtModel()
        elif peak_finding_model == "lorentzian":
            model = LorentzianModel()
        else:
            raise NotImplementedError(
                f"""
                {peak_finding_model} is not implemented! 
                Options: [gauss, voigt, lorentzian]
                """
            )

        fit_df = []
        fit_params = []
        fit_report = []
        for df in self.divided_peaks[assay_number]:
            df = df.copy()
            y = df.peaks.to_numpy()
            # CHANGED to time instead of basepair
            x = df.time.to_numpy()

            params = model.guess(y, x)
            out = model.fit(y, params, x=x)

            fit_df.append(df.assign(fitted=out.best_fit, model=peak_finding_model))
            fit_params.append(out.values)
            fit_report.append(out.fit_report())

        # Update the instances of the model fit
        self.fit_df = fit_df
        self.fit_params = fit_params
        self.fit_report = fit_report

    # TODO
    # Fix so that the cutoff is a range or something else.
    def calculate_quotient(
        self,
    ) -> None:

        """
        :Params:
        """
        areas = np.array([x["amplitude"] for x in self.fit_params])

        right_by_left = (
            self.cutoff is None
            or pd.concat(self.fit_df).basepairs.mean() >= self.cutoff
        )
        # if there only is 1 peak, return 0
        if len(areas) == 1:
            self.quotient = 0
            return

        # if there only are 2 peaks, return the quotient
        if len(areas) == 2:
            # left peak divided by right peak
            if not right_by_left:
                self.quotient = areas[0] / areas[1]
                return
            # right peak divided by left peak
            self.quotient = areas[1] / areas[0]
            return

        # return the last peak divided by the mean of the peaks to the left of it
        self.quotient = areas[-1] / areas[:-1].mean()

    def peak_position_area_dataframe(
        self, assay_number: int, name: str
    ) -> pd.DataFrame:
        """
        Returns a DataFrame of each peak and its properties
        """
        dataframes = []
        for i, _ in enumerate(self.fit_df):
            report = self.fit_report[i]
            r_value = float(re.findall(r"R-squared *= (0\.\d{3})", report)[0])
            df = (
                self.fit_df[i]
                .loc[lambda x: x.peaks == x.peaks.max()]
                .assign(area=self.fit_params[i]["amplitude"])
                .assign(r_value=r_value)
                .assign(peak_name=f"Peak {i + 1}")
                .drop(columns="time")
                .reset_index(drop=True)
                .rename(
                    columns={
                        "peaks": "peak_height",
                        "fitted": "fitted_peak_height",
                    }
                )
                .drop_duplicates("peak_name")
                .assign(file_name=self.file_name)
                .assign(assay_name=name)
            )
            dataframes.append(df)

        self.assay_peak_area_df = pd.concat(dataframes).assign(
            quotient=self.quotient,
            peak_number=lambda x: x.shape[0],
            assay_number=assay_number + 1,
        )

    def plot_areas(self, peak_finding_model: str):
        fig_areas, axs = plt.subplots(
            1, len(self.fit_df), sharey=True, figsize=(20, 10)
        )
        # if there is only one peak
        if len(self.fit_df) == 1:
            axs.plot(self.fit_df[0].basepairs, self.fit_df[0].peaks, "o")
            axs.plot(self.fit_df[0].basepairs, self.fit_df[0].fitted)
            axs.set_title(f"Peak 1 area: {self.fit_params[0]['amplitude']: .1f}")
            axs.grid()
        # if more than one peak
        else:
            for i, ax in enumerate(axs):
                ax.plot(
                    self.fit_df[i].basepairs,
                    self.fit_df[i].peaks,
                    "o",
                )
                ax.plot(self.fit_df[i].basepairs, self.fit_df[i].fitted)
                ax.set_title(
                    f"Peak {i + 1} area: {self.fit_params[i]['amplitude']: .1f}"
                )
                ax.grid()

        fig_areas.suptitle(f"Quotient: {self.quotient: .2f}")
        fig_areas.legend(["Raw data", "Model"])
        fig_areas.supxlabel("basepairs")
        fig_areas.supylabel("intensity")
        plt.close()

        return fig_areas

    def fit_assay_peaks(
        self,
        peak_finding_model: str,
        assay_number: int,
        name: str,
    ) -> None:
        """
        Runs fit_lmfit_model, calculate_quotient and peak_position_area_dataframe
        """
        self.fit_lmfit_model(peak_finding_model, assay_number)
        self.calculate_quotient()
        self.peak_position_area_dataframe(assay_number, name)
        # area plots
        area_plot = self.plot_areas(peak_finding_model)
        self.area_plots.append(area_plot)
        return self.assay_peak_area_df

    def assays_dataframe(self, peak_finding_model: str = "gauss"):
        dfs = []
        for i in self:
            if self.peaks.custom_peaks is not None:
                name = self.peaks.custom_peaks.name.unique()[i]
            else:
                name = ""
            dfs.append(self.fit_assay_peaks(peak_finding_model, i, name))
        df = pd.concat(dfs, ignore_index=True)
        # initialize attribute self.final_df
        self.final_df = df
        return df



from ..utils.peak_finder import PeakFinder
import matplotlib.pyplot as plt
import numpy as np


class PlotPeaks:
    def __init__(self, peaks: PeakFinder):
        self.peaks = peaks

    @property
    def plot_peaks(self):
        fig_peaks = plt.figure(figsize=(20, 10))

        df = self.peaks.peaks_dataframe.loc[
            lambda x: x.basepairs > self.peaks.peak_information.basepairs.min() - 10
        ].loc[lambda x: x.basepairs < self.peaks.peak_information.basepairs.max() + 10]

        plt.plot(df.basepairs, df.peaks)
        plt.plot(
            self.peaks.peak_information.basepairs,
            self.peaks.peak_information.peaks,
            "o",
        )
        for x, y in zip(
            self.peaks.peak_information.basepairs,
            self.peaks.peak_information.peaks,
        ):
            plt.text(x, y, f"{round(x, 1)} bp")

        channel = self.peaks.model.fsa_file.trace_channel
        plt.title(f"Channel: {channel}")
        plt.xticks(np.arange(df.basepairs.min(), df.basepairs.max(), 10), rotation=90)
        plt.ylabel("intensity")
        plt.xlabel("basepairs")
        plt.grid()
        plt.close()

        return fig_peaks



import fire
import pandas as pd
from pathlib import Path
import sys
import logging
from colorama import Fore as f, init
import matplotlib
import warnings
import platform

warnings.filterwarnings("ignore")

import fraggler

# from colorama
init(autoreset=True)

# for windows user
if platform.system() == "Windows":
    matplotlib.use("agg")

ASCII_ART = f"""{f.RED}
            █████▒██▀███   ▄▄▄        ▄████   ▄████  ██▓    ▓█████  ██▀███
          ▓██   ▒▓██ ▒ ██▒▒████▄     ██▒ ▀█▒ ██▒ ▀█▒▓██▒    ▓█   ▀ ▓██ ▒ ██▒
          ▒████ ░▓██ ░▄█ ▒▒██  ▀█▄  ▒██░▄▄▄░▒██░▄▄▄░▒██░    ▒███   ▓██ ░▄█ ▒
          ░▓█▒  ░▒██▀▀█▄  ░██▄▄▄▄██ ░▓█  ██▓░▓█  ██▓▒██░    ▒▓█  ▄ ▒██▀▀█▄
          ░▒█░   ░██▓ ▒██▒ ▓█   ▓██▒░▒▓███▀▒░▒▓███▀▒░██████▒░▒████▒░██▓ ▒██▒
           ▒ ░   ░ ▒▓ ░▒▓░ ▒▒   ▓▒█░ ░▒   ▒  ░▒   ▒ ░ ▒░▓  ░░░ ▒░ ░░ ▒▓ ░▒▓░
           ░       ░▒ ░ ▒░  ▒   ▒▒ ░  ░   ░   ░   ░ ░ ░ ▒  ░ ░ ░  ░  ░▒ ░ ▒░
           ░ ░     ░░   ░   ░   ▒   ░ ░   ░ ░ ░   ░   ░ ░      ░     ░░   ░
                    ░           ░  ░      ░       ░     ░  ░   ░  ░   ░
"""


def save_df_format(
    peak_dfs: list,
    out_folder: str,
    in_path: str,
    out_format: str,
) -> None:
    # Save dataframe
    df = pd.concat(peak_dfs).reset_index(drop=True)
    out_name = f"{out_folder}/areatable_{Path(in_path).resolve().parts[-1]}"
    if out_format == "excel":
        df.to_excel(f"{out_name}.xlsx", index=False)
    elif out_format == "csv":
        df.to_csv(f"{out_name}.csv", index=False)
    elif out_format == "json":
        df.to_json(f"{out_name}.json")
    else:
        raise NotImplementedError("Choose between: csv, excel, json")


def area_report(
    in_path: str,
    out_folder: str,
    ladder: str,
    peak_model: str = "gauss",
    min_ratio: float = 0.2,
    min_height: int = 30,
    cutoff: int = 175,
    channel: str = "DATA1",
    peak_height: int = 500,
    size_standard_channel: str | None = None,
    distance_between_assays: int = 15,
    custom_peaks: str = None,
    out_format: str = "excel",
) -> None:

    print(ASCII_ART)
    if custom_peaks:
        peak_height = 0

    # Logging
    fraggler.setup_logging(out_folder)
    INFO = f"""
    Runned command:
    {' '.join(sys.argv)}

    Running fraggler with following parameters:
        In path: {in_path}
        Out folder: {out_folder}
        Ladder: {ladder}
        Peak model: {peak_model}
        Min ratio: {min_ratio}
        Min height: {min_height}
        Cutoff: {cutoff}
        Trace channel: {channel}
        Peak Height: {peak_height}
        Custom Peaks: {custom_peaks}
        Out format: {out_format}
        Size standard channel: {size_standard_channel}
        Distance between assays: {distance_between_assays}
    """
    logging.info(INFO)

    # Files
    files = fraggler.get_files(in_path)
    out_folder = Path(out_folder)

    # Generate a peak area report for each file
    failed_files = []
    no_peaks = []
    peak_dfs = []
    for file in files:
        logging.info(f"Processing file: {file}")
        fraggler_object = fraggler.make_fraggler_area(
            file=file,
            ladder=ladder,
            min_height=min_height,
            min_ratio=min_ratio,
            trace_channel=channel,
            distance_between_assays=distance_between_assays,
            size_standard_channel=size_standard_channel,
            cutoff=cutoff,
            peak_height=peak_height,
            custom_peaks=custom_peaks,
        )

        # If make_fraggler_object failed
        if isinstance(fraggler_object, str):
            failed_files.append(fraggler_object)
            fsa = fraggler.FsaFile(file, ladder)
            report = fraggler.generate_no_peaks_report(fsa)
            out_name = out_folder / f"{file.stem}_fraggler_failed.html"
            report.save(out_name)
            continue

        if not fraggler_object.peaks.found_peaks:
            no_peaks.append(file.stem)
            fsa = fraggler.FsaFile(file, ladder)
            report = fraggler.generate_no_peaks_report(fsa)
            out_name = out_folder / f"{file.stem}_failed.html"
            report.save(out_name)
            continue

        # generate report and peak table

        peak_dfs.append(fraggler_object.areas.assays_dataframe(peak_model))
        report = fraggler.generate_area_report(fraggler_object, peak_model)
        out_name = out_folder / f"{file.stem}_fraggler_area.html"
        report.save(out_name)

    # Save dataframe
    if peak_dfs:
        save_df_format(peak_dfs, out_folder, in_path, out_format)

    # log failed files
    logging.info(f"Fraggler done for files in {in_path}!")

    if failed_files:
        failed_files = "\n".join(failed_files)
        logging.warning(f"{f.YELLOW}Following files failed: {failed_files}")
    if no_peaks:
        no_peaks = "\n".join(no_peaks)
        logging.warning(f"{f.YELLOW}Following files had no peaks: {no_peaks}")


def peak_report(
    in_path: str,
    out_folder: str,
    ladder: str,
    min_ratio: float = 0.2,
    min_height: int = 30,
    channel: str = "DATA1",
    peak_height: int = 500,
    size_standard_channel: str | None = None,
    distance_between_assays: int = 15,
    custom_peaks: str = None,
    out_format: str = "excel",
) -> None:

    if custom_peaks:
        peak_height = 0

    print(ASCII_ART)

    # Logging
    fraggler.setup_logging(out_folder)
    INFO = f"""
    Runned command:
    {' '.join(sys.argv)}

    Running fraggler with following parameters:
        In path: {in_path}
        Out folder: {out_folder}
        Ladder: {ladder}
        Min ratio: {min_ratio}
        Min height: {min_height}
        Trace channel: {channel}
        Peak Height: {peak_height}
        Custom Peaks: {custom_peaks}
        Out format: {out_format}
        Size standard channel: {size_standard_channel}
        Distance between assays: {distance_between_assays}
    """
    logging.info(INFO)

    # Files
    files = fraggler.get_files(in_path)
    out_folder = Path(out_folder)

    # Generate a peak area report for each file
    failed_files = []
    no_peaks = []
    peak_dfs = []
    for file in files:
        logging.info(f"Processing file: {file}")
        fraggler_object = fraggler.make_fraggler_peak(
            file=file,
            ladder=ladder,
            min_height=min_height,
            min_ratio=min_ratio,
            trace_channel=channel,
            distance_between_assays=distance_between_assays,
            size_standard_channel=size_standard_channel,
            peak_height=peak_height,
            custom_peaks=custom_peaks,
        )

        # If make_fraggler_object failed
        if isinstance(fraggler_object, str):
            failed_files.append(fraggler_object)
            fsa = fraggler.FsaFile(file, ladder)
            report = fraggler.generate_no_peaks_report(fsa)
            out_name = out_folder / f"{file.stem}_fraggler_failed.html"
            report.save(out_name)
            continue

        if not fraggler_object.peaks.found_peaks:
            no_peaks.append(file.stem)
            fsa = fraggler.FsaFile(file, ladder)
            report = fraggler.generate_no_peaks_report(fsa)
            out_name = out_folder / f"{file.stem}_failed.html"
            report.save(out_name)
            continue

        # add peaks to dataframe
        df = fraggler_object.peaks.peak_information.assign(
            file_name=fraggler_object.fsa.file_name
        )[["file_name", "basepairs", "peaks"]].rename(columns={"peaks": "peak_height"})
        peak_dfs.append(df)

        # generate report
        report = fraggler.generate_peak_report(fraggler_object)
        out_name = out_folder / f"{file.stem}_fraggler_peak.html"
        report.save(out_name)

    # Save dataframe
    if peak_dfs:
        save_df_format(peak_dfs, out_folder, in_path, out_format)

    # log failed files
    logging.info(f"Fraggler done for files in {in_path}!")

    if failed_files:
        failed_files = "\n".join(failed_files)
        logging.warning(f"{f.YELLOW}Following files failed: {failed_files}")
    if no_peaks:
        no_peaks = "\n".join(no_peaks)
        logging.warning(f"{f.YELLOW}Following files had no peaks: {no_peaks}")


def run():
    """
    Run the command-line interface using Fire.
    """
    fire.Fire(
        {
            "area": area_report,
            "peak": peak_report,
        }
    )



import pandas as pd
import numpy as np
import networkx as nx
from scipy import signal
from sklearn.metrics import r2_score
from sklearn.preprocessing import minmax_scale
from scipy.interpolate import UnivariateSpline

from fraggler.utils.fsa_file import FsaFile


class PeakLadderAssigner:
    """
    A class that assigns peak sizes to a ladder for use in DNA fragment size analysis.
    The peak sizes are generated based on a given size standard and the input parameters.

    Attributes:
    fsa_file (FsaFile): an object of the FsaFile class that contains parameters for analysis

    Methods:
    assign_ladder_peak_sizes(): Assigns peak sizes to a ladder based on the given parameters and returns the best combination.
    get_peaks(size_standard): Finds peaks in the size standard based on given parameters and returns the peaks.
    generate_graph(peaks): Generates a graph based on the peaks and returns the graph.
    generate_combinations(graph): Generates combinations of peaks based on the graph and returns the combinations.
    get_best_fit(combinations, method): Finds the best fit for the given combinations and returns the best combination.
    _polynomial_model_inv_r2_score(ladder, comb): Returns a score based on the polynomial model inverse r2 score.
    _max_fractional_deviation_score(ladder, comb): Returns a score based on the maximum fractional deviation.
    _max_first_derivative_score(combination): Returns a score based on the maximum first derivative.
    _max_second_derivative_score(combination): Returns a score based on the maximum second derivative.
    _max_spline_second_derivative_score(combination): Returns a score based on the maximum spline second derivative.
    """

    def __init__(self, fsa_file: FsaFile) -> None:
        """
        Initializes the class with an object of the FsaFile class.

        Args:
        fsa_file (FsaFile): an object of the FsaFile class that contains parameters for analysis
        """

        self.fsa_file = fsa_file

    def assign_ladder_peak_sizes(self):
        """
        Assigns peak sizes to a ladder based on the given parameters and returns the best combination.

        Returns:
        np.array: the best combination of peak sizes
        """
        peaks = self.get_peaks(self.fsa_file.size_standard)
        graph = self.generate_graph(peaks)
        combinations = self.generate_combinations(graph)
        best_combination = self.get_best_fit(combinations)

        return best_combination

    def get_peaks(self, size_standard) -> np.array:
        """
        Finds peaks in the size standard based on given parameters and returns the peaks.

        Args:
        size_standard (np.array): the size standard array

        Returns:
        np.array: the peaks array
        """

        peaks_obj = signal.find_peaks(
            size_standard,
            distance=self.fsa_file.min_interpeak_distance,
            height=self.fsa_file.min_height,
        )

        peaks = peaks_obj[0]
        heights = peaks_obj[1]["peak_heights"]

        df = pd.DataFrame({"peaks": peaks, "heights": heights})

        # TODO
        # Dropped the below, then it worked with the new test files
        # idxmax = df["heights"].idxmax()
        # df = df.drop(idxmax)

        peaks_adj = df.nlargest(self.fsa_file.max_peak_count, ["heights"])

        return peaks_adj["peaks"].sort_values().to_numpy()

    def generate_graph(self, peaks) -> nx.DiGraph:
        """
        Generates a graph based on the peaks and returns the graph.

        Args:
        peaks (np.array): the peaks array

        Returns:
        nx.DiGraph: the graph object
        """
        G = nx.DiGraph()

        for p in peaks:
            G.add_node(p)

        i = 0
        while i < peaks.size:
            j = i + 1
            while j < peaks.size:
                diff = peaks[j] - peaks[i]
                if diff <= self.fsa_file.max_ladder_trace_distance:
                    G.add_edge(peaks[i], peaks[j], length=diff)
                j += 1
            i += 1

        return G

    def generate_combinations(self, graph):
        """
        Generates combinations of nodes from a given directed acyclic graph (DAG)
        that satisfy certain conditions.

        Parameters:
        graph (networkx.DiGraph): a DAG representing a state machine

        Returns:
        A numpy array representing a sequence of nodes that satisfies the
        conditions of the DAG.
        """

        # Get start nodes that have zero in-degree
        start_nodes = [node for node in graph.nodes if graph.in_degree(node) == 0]

        # Get end nodes that have zero out-degree
        end_nodes = [node for node in graph.nodes if graph.out_degree(node) == 0]
        if len(start_nodes) == 0 or len(end_nodes) == 0:
            raise ValueError("Graph does not have start or end nodes")

        # Get all simple paths from start node to end node
        all_paths = []
        for start_node in start_nodes:
            for end_node in end_nodes:
                paths = nx.all_simple_paths(graph, start_node, end_node)
                all_paths.extend(paths)

        if len(all_paths) == 0:
            raise ValueError("No paths found from start to end nodes")

        # Generate combinations of nodes that satisfy certain conditions
        for p_arr in all_paths:
            for i in range(0, len(p_arr) - self.fsa_file.ref_count + 1):
                yield np.array(p_arr[i : i + self.fsa_file.ref_count])

    def get_best_fit(self, combinations, method="2nd_derivative"):
        """
        Calculates the best fit for a given set of combinations using a specified method.

        Parameters:
        combinations (iterable): an iterable containing combinations of nodes
        method (str): a string indicating the method to use. The default method is
            "2nd_derivative".

        Returns:
        A numpy array representing the combination of nodes with the best fit.
        """

        # Create an empty dataframe to store combinations and scores
        df = pd.DataFrame()

        # Store the combinations in the dataframe
        df["combination"] = list(combinations)

        # Calculate the score for each combination using the specified method
        if method == "2nd_derivative":
            df["score"] = np.vectorize(self._max_spline_second_derivative_score)(
                df["combination"]
            )

        if method == "first_derivative":
            df["score"] = np.vectorize(self._max_first_derivative_score)(
                df["combination"]
            )

        # Sort the dataframe by score in ascending order
        df = df.sort_values(by="score", ascending=True)

        # Get the best combination (i.e., the one with the lowest score)
        best = df.head(1)

        # Return the best combination as a numpy array
        return best.combination.squeeze()

    @staticmethod
    def _polynomial_model_inv_r2_score(ladder: np.array, comb: np.array) -> float:
        """
        Calculates the inverse of the R^2 (coefficient of determination) score for a polynomial regression model.

        Parameters:
        - ladder (np.array): array containing the ladder positions of a dataset
        - comb (np.array): array containing the corresponding comb heights of a dataset

        Returns:
        - float: the inverse of the R^2 score for the polynomial model

        Notes:
        - This method fits a polynomial regression model of degree 3 to the dataset using NumPy's polyfit function.
        - It then calculates the predicted values of the comb heights using NumPy's poly1d function.
        - Finally, it calculates the R^2 score between the predicted comb heights and the actual comb heights using the r2_score function
          from the scikit-learn library, and returns the inverse of that score.

        """
        # fit polynomial regression model
        fit = np.polyfit(ladder, comb, 3)

        # create predicted values using fitted model
        predict = np.poly1d(fit)

        # calculate R^2 score between predicted and actual values, and return its inverse
        return 1 - r2_score(comb, predict(ladder))

    @staticmethod
    def _max_fractional_deviation_score(ladder: np.ndarray, comb: np.ndarray):
        """
        Calculates the maximum fractional deviation score for a dataset.

        Parameters:
        - ladder (np.ndarray): array containing the ladder positions of a dataset
        - comb (np.ndarray): array containing the corresponding comb heights of a dataset

        Returns:
        - float: the maximum fractional deviation score

        Notes:
        - This method calculates the interval lengths between adjacent ladder positions and corresponding comb heights using NumPy's
          diff function and min-max scaling with a feature range equal to the range of ladder positions.
        - It then calculates the fractional deviation between the interval lengths and the scaled comb heights, and returns the
          maximum deviation.

        """
        # calculate interval lengths and scaled comb height intervals
        l_intervals = np.diff(ladder)
        c_intervals = np.diff(minmax_scale(comb, feature_range=(ladder[0], ladder[-1])))

        # calculate fractional deviation between interval lengths and scaled comb height intervals
        frac_deviation = np.abs(l_intervals - c_intervals) / l_intervals

        # find index of maximum deviation and return the corresponding value
        max_frac_deviation_idx = np.argmax(frac_deviation)
        return frac_deviation[max_frac_deviation_idx]

    def _max_first_derivative_score(self, combination: np.ndarray) -> float:
        """
        Calculates the maximum first derivative score for the given combination of features.

        Args:
            combination (np.ndarray): A 1-dimensional numpy array containing a combination of features.

        Returns:
            float: The maximum first derivative score for the given combination of features.
        """
        # Scale the combination to the range of reference sizes
        comb_scaled = minmax_scale(
            combination,
            feature_range=(self.fsa_file.ref_sizes[0], self.fsa_file.ref_sizes[-1]),
        )

        # Calculate the differences between consecutive values of the scaled combination and the reference sizes
        diff_intervals = np.diff(comb_scaled) - np.diff(self.fsa_file.ref_sizes)

        # Calculate the absolute gradient of the differences between consecutive values
        abs_diff_intervals_gradient = np.abs(np.gradient(diff_intervals))

        # Find the index of the maximum absolute gradient
        max_abs_diff_intervals_gradient_idx = np.argmax(abs_diff_intervals_gradient)

        # Return the maximum absolute gradient
        return abs_diff_intervals_gradient[max_abs_diff_intervals_gradient_idx]

    def _max_second_derivative_score(self, combination: np.ndarray) -> float:
        """
        Calculates the maximum absolute second derivative score of the given combination.

        Args:
            combination (np.ndarray): An array of feature values for a given combination.

        Returns:
            float: The maximum absolute second derivative score of the given combination.

        Raises:
            None.

        """
        # Scale the combination to match the range of reference sizes.
        comb_scaled = minmax_scale(
            combination,
            feature_range=(self.fsa_file.ref_sizes[0], self.fsa_file.ref_sizes[-1]),
        )

        # Calculate the difference between the scaled combination and reference sizes.
        diff_intervals = np.diff(comb_scaled) - np.diff(self.fsa_file.ref_sizes)

        # Calculate the absolute second derivative of the difference intervals.
        abs_second_derivative = np.abs(np.gradient(np.gradient(diff_intervals)))

        # Find the index of the maximum absolute second derivative score.
        max_second_derivative_idx = np.argmax(abs_second_derivative)

        # Return the maximum absolute second derivative score.
        return abs_second_derivative[max_second_derivative_idx]

    def _max_spline_second_derivative_score(self, combination: np.ndarray) -> float:
        """
        Calculates the maximum absolute value of the second derivative of a given combination of values using a UnivariateSpline.

        Args:
        - self (object): an instance of the class containing the function
        - combination (np.ndarray): an array of values representing a combination

        Returns:
        - max_value (float): the maximum absolute value of the second derivative of the given combination

        """
        # create a UnivariateSpline object using the reference sizes and the given combination
        spl = UnivariateSpline(self.fsa_file.ref_sizes, combination, s=0)

        # calculate the second derivative of the spline
        der2 = spl.derivative(n=2)

        # evaluate the second derivative at each reference size and return the maximum absolute value
        max_value = max(abs(der2(self.fsa_file.ref_sizes)))

        return max_value



import panel as pn
from pathlib import Path
import datetime
import pandas as pd

from ..utils.peak_finder import PeakFinder
from ..utils.fraggler_object import (
    FragglerPeak,
    FragglerArea,
    make_fraggler_peak,
    make_fraggler_area,
)
from ..utils.fsa_file import FsaFile
from ..plotting.plot_channels import plot_fsa_data
from ..plotting.plot_peaks import PlotPeaks
from ..plotting.plot_ladder import PlotLadder


pn.extension("tabulator")
pn.extension("vega", sizing_mode="stretch_width", template="fast")
pn.widgets.Tabulator.theme = "modern"


def header(
    text: str,
    bg_color: str = "#04c273",
    height: int = 300,
    fontsize: str = "px20",
    textalign: str = "center",
):
    """
    Template for markdown header like block
    """
    return pn.pane.Markdown(
        f"""
        {text}
        """,
        background=bg_color,
        height=height,
        margin=10,
        style={
            "color": "white",
            "padding": "10px",
            "text-align": f"{textalign}",
            "font-size": f"{fontsize}",
        },
    )


def make_header(name: str, date: str) -> pn.pane.Markdown:
    return header(
        text=f"""
        # Fragment Analysis Report
        ## Report of {name}
        ## Date: {date}
        """,
        fontsize="20px",
        bg_color="#03a1fc",
        height=250,
    )


def generate_peak_report(fraggler: FragglerPeak) -> pn.layout.base.Column:
    ### ----- Raw Data ----- ###
    channel_header = header(
        text="## Plot of channels",
        bg_color="#04c273",
        height=80,
        textalign="left",
    )
    # PLOT
    channel_tab = pn.Tabs()
    for plot, name in plot_fsa_data(fraggler.fsa):
        pane = pn.pane.Vega(plot.interactive(), sizing_mode="stretch_both", name=name)
        channel_tab.append(pane)

    channels_section = pn.Column(channel_header, channel_tab)

    ### ----- Peaks ----- ###
    peaks_header = header(
        text="## Plot of Peaks",
        bg_color="#04c273",
        height=80,
        textalign="left",
    )

    # PLOT
    peaks_plot = PlotPeaks(fraggler.peaks).plot_peaks
    peaks_pane = pn.pane.Matplotlib(peaks_plot, name="Peaks")

    # Section
    peaks_tab = pn.Tabs(
        peaks_pane,
    )
    peaks_section = pn.Column(peaks_header, peaks_tab)

    ### ----- Ladder Information ----- ###
    ladder_header = header(
        text="## Information about the ladder",
        bg_color="#04c273",
        height=80,
        textalign="left",
    )
    # Ladder peak plot
    ladder_plot = PlotLadder(fraggler.model)
    ladder_peak_plot = pn.pane.Matplotlib(
        ladder_plot.plot_ladder_peaks,
        name="Ladder Peak Plot",
    )
    # Ladder Correlation
    ladder_correlation_plot = pn.pane.Matplotlib(
        ladder_plot.plot_model_fit,
        name="Ladder Correlation Plot",
    )

    # Section
    ladder_tab = pn.Tabs(
        ladder_peak_plot,
        ladder_correlation_plot,
    )
    ladder_section = pn.Column(ladder_header, ladder_tab)

    ### ----- Peaks dataframe ----- ###
    dataframe_header = header(
        text="## Peaks Table", bg_color="#04c273", height=80, textalign="left"
    )
    # Create dataframe
    df = fraggler.peaks.peak_information.assign(file_name=fraggler.fsa.file_name)[
        ["file_name", "basepairs", "peaks", "assay_name"]
    ].rename(columns={"peaks": "peak_height"})
    # DataFrame Tabulator
    peaks_df_tab = pn.widgets.Tabulator(
        df,
        layout="fit_columns",
        pagination="local",
        page_size=15,
        show_index=False,
        name="Peaks Table",
    )

    # Section
    dataframe_tab = pn.Tabs(peaks_df_tab)
    dataframe_section = pn.Column(dataframe_header, dataframe_tab)

    ### CREATE REPORT ###

    file_name = fraggler.fsa.file_name
    date = fraggler.fsa.fsa["RUND1"]
    head = make_header(file_name, date)

    all_tabs = pn.Tabs(
        ("Channels", channels_section),
        ("Peaks", peaks_section),
        ("Ladder", ladder_section),
        ("Peaks Table", dataframe_section),
        tabs_location="left",
    )
    report = pn.Column(
        head,
        pn.layout.Divider(),
        all_tabs,
    )

    return report


def generate_area_report(
    fraggler: FragglerArea,
    peak_model: str = "gauss",
) -> pn.layout.base.Column:

    ### ----- Raw Data ----- ###
    channel_header = header(
        text="## Plot of channels",
        bg_color="#04c273",
        height=80,
        textalign="left",
    )
    # PLOT
    channel_tab = pn.Tabs()
    for plot, name in plot_fsa_data(fraggler.fsa):
        pane = pn.pane.Vega(plot.interactive(), sizing_mode="stretch_both", name=name)
        channel_tab.append(pane)

    channels_section = pn.Column(channel_header, channel_tab)

    ### ----- Peaks ----- ###
    peaks_header = header(
        text="## Plot of Peaks",
        bg_color="#04c273",
        height=80,
        textalign="left",
    )

    # PLOT
    peaks_plot = PlotPeaks(fraggler.peaks).plot_peaks
    peaks_pane = pn.pane.Matplotlib(peaks_plot, name="Peaks")

    # Section
    peaks_tab = pn.Tabs(
        peaks_pane,
    )
    peaks_section = pn.Column(peaks_header, peaks_tab)

    ### ----- Ladder Information ----- ###
    ladder_header = header(
        text="## Information about the ladder",
        bg_color="#04c273",
        height=80,
        textalign="left",
    )
    # Ladder peak plot
    ladder_plot = PlotLadder(fraggler.model)
    ladder_peak_plot = pn.pane.Matplotlib(
        ladder_plot.plot_ladder_peaks,
        name="Ladder Peak Plot",
    )
    # Ladder Correlation
    ladder_correlation_plot = pn.pane.Matplotlib(
        ladder_plot.plot_model_fit,
        name="Ladder Correlation Plot",
    )

    # Section
    ladder_tab = pn.Tabs(
        ladder_peak_plot,
        ladder_correlation_plot,
    )
    ladder_section = pn.Column(ladder_header, ladder_tab)

    ### ----- Areas Information ----- ###
    areas_header = header(
        text="## Peak Areas", bg_color="#04c273", height=80, textalign="left"
    )
    areas_tab = pn.Tabs()
    for i, plot in enumerate(fraggler.areas.area_plots):
        name = f"Assay {i + 1}"
        plot_pane = pn.pane.Matplotlib(plot, name=name)
        areas_tab.append(plot_pane)

    # Section
    areas_section = pn.Column(areas_header, areas_tab)

    ### ----- Peaks DataFrame ----- ###
    dataframe_header = header(
        text="## Peaks Table", bg_color="#04c273", height=80, textalign="left"
    )

    if not hasattr(fraggler.areas, "final_df"):
        df = fraggler.areas.assays_dataframe(peak_model)
    else:
        df = fraggler.areas.final_df

    # DataFrame Tabulator
    peaks_df_tab = pn.widgets.Tabulator(
        df,
        layout="fit_columns",
        pagination="local",
        page_size=15,
        show_index=False,
        name="Peaks Table",
    )

    # Section
    dataframe_tab = pn.Tabs(peaks_df_tab)
    dataframe_section = pn.Column(dataframe_header, dataframe_tab)

    ### CREATE REPORT ###

    file_name = fraggler.fsa.file_name
    date = fraggler.fsa.fsa["RUND1"]
    head = make_header(file_name, date)

    all_tabs = pn.Tabs(
        ("Channels", channels_section),
        ("Peaks", peaks_section),
        ("Ladder", ladder_section),
        ("Areas", areas_section),
        ("Peak Table", dataframe_section),
        tabs_location="left",
    )
    report = pn.Column(
        head,
        pn.layout.Divider(),
        all_tabs,
    )

    return report


def generate_no_peaks_report(fsa: FsaFile):
    channel_header = header(
        text="## Plot of channels",
        bg_color="#04c273",
        height=80,
        textalign="left",
    )
    # PLOT
    channel_tab = pn.Tabs()
    for plot, name in plot_fsa_data(fsa):
        pane = pn.pane.Vega(plot.interactive(), sizing_mode="stretch_both", name=name)
        channel_tab.append(pane)
    channels_section = pn.Column(channel_header, channel_tab)

    ### CREATE REPORT ###
    file_name = fsa.file_name
    date = fsa.fsa["RUND1"]
    head = header(
        "# No peaks could be generated. Please look at the raw data.", height=100
    )

    all_tabs = pn.Tabs(
        ("Channels", channels_section),
        tabs_location="left",
    )
    report = pn.Column(
        head,
        pn.layout.Divider(),
        all_tabs,
    )

    return report



"""
Class to fit model to ladder and size standard
"""

import logging
from sklearn.pipeline import make_pipeline
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import SplineTransformer
from sklearn.metrics import mean_squared_error, r2_score
import pandas as pd

from fraggler.ladder_fitting.peak_ladder_assigner import PeakLadderAssigner
from fraggler.utils.fsa_file import FsaFile


class ModelFittingError(Exception):
    pass


class LadderNotFoundError(Exception):
    pass


class FitLadderModel:
    def __init__(self, ladder_assigner: PeakLadderAssigner) -> None:
        """
        Initialize FitLadderModel object with ladder_assigner object and its attributes.

        Args:
            ladder_assigner (PeakLadderAssigner): Object of PeakLadderAssigner class.

        Returns:
            None.
        """
        self.n_knots = 2
        self.ladder_assigner = ladder_assigner
        self.fsa_file = self.ladder_assigner.fsa_file
        self.best_combination = ladder_assigner.assign_ladder_peak_sizes().reshape(
            -1, 1
        )

        self.fit_model()
        self.mse, self.r2 = self.model_score()
        self.adjusted_baisepair_df = self.generate_adjusted_trace_df()

    def fit_model(self) -> None:
        """
        Fit model based on ladder type.

        Returns:
            None.
        """
        ladder = self.fsa_file.ladder
        if ladder == "ROX":
            self._fit_ROX_ladder()
        elif ladder == "LIZ" or ladder == "ORANGE":
            self._fit_LIZ_ladder()
        else:
            raise LadderNotFoundError(f"'{ladder}' is not a valid ladder")

    def model_score(self) -> tuple[float, float]:
        """
        Calculate mean squared error and R-squared score of the fitted model.

        Returns:
            Tuple (mse, r2).
        """
        true_Y = self.fsa_file.ref_sizes
        predicted_Y = self.model.predict(self.best_combination)
        mse = mean_squared_error(true_Y, predicted_Y)
        r2 = r2_score(true_Y, predicted_Y)

        return mse, r2

    def generate_adjusted_trace_df(self) -> pd.DataFrame:
        """
        Generate a dataframe with adjusted basepairs and peaks.

        Returns:
            Dataframe with columns time, peaks and basepairs.
        """
        # Continue loop until all basepairs are unique by changing n_knots in SplineTransformer
        for _ in range(10):
            df = (
                pd.DataFrame({"peaks": self.fsa_file.trace})
                .reset_index()
                .rename(columns={"index": "time"})
                .assign(
                    basepairs=lambda x: self.model.predict(
                        x.time.to_numpy().reshape(-1, 1)
                    )
                )
                .loc[lambda x: x.basepairs >= 0]
            )

            if df.shape[0] == df.basepairs.nunique():
                logging.info(f"Ladder fitting model: {self.model}")
                return df
            # If not all bp are unique
            self.n_knots += 1
            self.fit_model()

        # reaches only if there is a problem
        raise ModelFittingError(
            "There is a problem with the fitting of the model to the ladder"
        )

    def _fit_ROX_ladder(self) -> None:
        """
        Fit model with ROX ladder.

        Returns:
            None.
        """
        self.model = make_pipeline(
            SplineTransformer(
                degree=4, n_knots=self.n_knots + 4, extrapolation="continue"
            ),
            LinearRegression(fit_intercept=True),
        )

        X = self.best_combination
        y = self.fsa_file.ref_sizes

        self.model.fit(X, y)

    def _fit_LIZ_ladder(self) -> None:
        """
        Fit model with LIZ ladder.

        Returns:
            None.
        """
        # NB!
        # Changed the SplineTransformer(n_knots=2) instead of 3!!
        self.model = make_pipeline(
            SplineTransformer(degree=3, n_knots=self.n_knots, extrapolation="continue"),
            LinearRegression(fit_intercept=True),
        )

        X = self.best_combination
        y = self.fsa_file.ref_sizes

        self.model.fit(X, y)



import matplotlib.pyplot as plt
import matplotlib
import numpy as np

from fraggler.ladder_fitting.fit_ladder_model import FitLadderModel


class PlotLadder:
    def __init__(self, model: FitLadderModel) -> None:
        self.model = model

    @property
    def plot_ladder_peaks(self) -> matplotlib.figure.Figure:
        trace = self.model.fsa_file.size_standard
        best_combination = self.model.best_combination
        ladder_name = self.model.fsa_file.ladder
        ladder_size = self.model.fsa_file.ref_sizes

        fig_ladder_peaks = plt.figure(figsize=(20, 10))
        plt.plot(trace)
        plt.plot(best_combination, trace[best_combination], "o")
        plt.xlabel("Time")
        plt.ylabel("Intensity")
        plt.legend(["Trace", "Peak (bp)"])
        plt.title(ladder_name)
        plt.grid()

        for peak, ladder in zip(best_combination, ladder_size):
            plt.text(peak, trace[peak], ladder)

        plt.close()
        return fig_ladder_peaks

    @property
    def plot_model_fit(self):

        ladder_size = self.model.fsa_file.ref_sizes
        best_combination = self.model.best_combination

        predicted = self.model.model.predict(best_combination)
        ladder_name = self.model.fsa_file.ladder

        mse = self.model.mse
        r2 = self.model.r2

        fig_model_fit = plt.figure(figsize=(20, 10))
        plt.plot(ladder_size, best_combination, "o")
        plt.plot(predicted, best_combination, "x")
        plt.xticks(np.arange(0, np.max(ladder_size), 50))
        plt.xlabel("bp")
        plt.yticks(np.arange(0, np.max(best_combination), 500))
        plt.suptitle(ladder_name)
        plt.title(f"{mse=}, {r2=}")
        # plt.title(f"{self.model.model[0]=}")
        plt.legend(["True value", "Predicted value"])
        plt.grid()

        plt.close()
        return fig_model_fit



"""
Taken from:
https://stackoverflow.com/questions/29156532/python-baseline-correction-library

from this paper:
https://pubs.rsc.org/en/content/articlelanding/2015/AN/C4AN01061B#!divAbstract
"""

from scipy import sparse
from scipy.sparse import linalg
import numpy as np
from numpy.linalg import norm


def baseline_arPLS(y, ratio=0.99, lam=100, niter=1000, full_output=False):
    L = len(y)

    diag = np.ones(L - 2)
    D = sparse.spdiags([diag, -2 * diag, diag], [0, -1, -2], L, L - 2)

    H = lam * D.dot(D.T)  # The transposes are flipped w.r.t the Algorithm on pg. 252

    w = np.ones(L)
    W = sparse.spdiags(w, 0, L, L)

    crit = 1
    count = 0

    while crit > ratio:
        z = linalg.spsolve(W + H, W * y)
        d = y - z
        dn = d[d < 0]

        m = np.mean(dn)
        s = np.std(dn)

        w_new = 1 / (1 + np.exp(2 * (d - (2 * s - m)) / s))

        crit = norm(w_new - w) / norm(w)

        w = w_new
        W.setdiag(w)  # Do not create a new matrix, just update diagonal values

        count += 1

        if count > niter:
            print("Maximum number of iterations exceeded")
            break

    if full_output:
        info = {"num_iter": count, "stop_criterion": crit}
        return z, d, info
    else:
        return z



import matplotlib.pyplot as plt
import fraggler


class PlotRawData:
    def __init__(self, fsa: fraggler.FsaFile):
        self.fsa = fsa

    @property
    def plot_raw_data(self):
        fig = plt.figure(figsize=(20, 10))

        plt.plot(self.fsa.trace)
        plt.xlabel("Time")
        plt.ylabel("Intensity")

        plt.close()
        return fig



from pathlib import Path
from Bio import SeqIO
import numpy as np

from fraggler.ladders.ladders import LADDERS
from fraggler.utils.baseline_removal import baseline_arPLS


class LadderNotFoundError(Exception):
    pass


class FsaFile:
    def __init__(
        self,
        file: str,
        ladder: str,
        normalize: bool = False,
        trace_channel: str = "DATA1",
        size_standard_channel: str = None,
        min_interpeak_distance: int = None,
        min_height: int = None,
        max_ladder_trace_distance: int = None,
    ) -> None:
        """
        Constructs an FsaFile object with the given parameters.

        Args:
            file (str): The path to the sequencing file.
            ladder (str): The name of the ladder used for sequencing.
            normalize (bool): Whether to normalize the data using the arPLS algorithm.
            trace_channel (str): The channel to extract the trace data from.
            size_standard_channel (str): The channel to extract the size standard data from.
            min_interpeak_distance (int): The minimum distance between peaks in the ladder.
            min_height (int): The minimum height of peaks in the ladder.
            max_ladder_trace_distance (int): The maximum distance between the ladder and the trace.

        Returns:
            None
        """
        peak_count_padding = 3
        self.file = Path(file)
        self.file_name = self.file.parts[-1]

        # Extract data from the sequencing file
        if ladder not in LADDERS.keys():
            raise LadderNotFoundError(f"'{ladder}' is not a valid ladder")
        self.ladder = ladder
        self.fsa = SeqIO.read(file, "abi").annotations["abif_raw"]
        self.trace_channel = trace_channel
        self.normalize = normalize

        # Extract data from the ladder reference
        self.ref_sizes = LADDERS[ladder]["sizes"]
        self.ref_count = self.ref_sizes.size

        # Use default values if nothing is given
        self.size_standard_channel = size_standard_channel or LADDERS[ladder]["channel"]
        self.min_interpeak_distance = (
            min_interpeak_distance or LADDERS[ladder]["distance"]
        )
        self.min_height = min_height or LADDERS[ladder]["height"]
        self.max_ladder_trace_distance = (
            max_ladder_trace_distance or LADDERS[ladder]["max_ladder_trace_distance"]
        )
        self.max_peak_count = self.ref_count + peak_count_padding

        # Normalize data if requested
        if normalize:
            self.size_standard = np.array(
                baseline_arPLS(self.fsa[self.size_standard_channel])
            )
            self.trace = np.array(baseline_arPLS(self.fsa[trace_channel]))
        else:
            self.size_standard = np.array(self.fsa[self.size_standard_channel])
            self.trace = np.array(self.fsa[trace_channel])

    def __str__(self):
        """
        Returns a string representation of the FsaFile object.

        Returns:
            str: A string representation of the object.
        """
        return f"""
            FsaFile-object with following parameters:
            
            File: {self.file}
            Filename: {self.file_name}
            Size standard channel: {self.size_standard_channel}
            Ladder name: {self.ladder}
            Number of ladder steps: {self.ref_count}
            Minimum interpeak distance: {self.min_interpeak_distance}
            Minimum height: {self.min_height}
            Minimum Ladder trace distance: {self.max_ladder_trace_distance}
            Maximum peak count: {self.max_peak_count}
            Normalized data: {self.normalize}
            Trace Channel: {self.trace_channel}
            Ladder Sizes: {self.ref_sizes}
            """



import setuptools
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setuptools.setup(
    name="fraggler",
    version="2.0.1",
    description="Fragment Analysis package in python!",
    url="https://github.com/willros/fraggler",
    author="William Rosenbaum and Pär Larsson",
    author_email="william.rosenbaum@umu.se",
    license="MIT",
    packages=setuptools.find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.10",
    install_requires=[
        "pandas",
        "numpy",
        "scikit-learn",
        "matplotlib",
        "networkx",
        "lmfit",
        "scipy",
        "biopython",
        "panel==1.3.8",
        "bokeh==3.3.4",
        "fire",
        "openpyxl",
        "colorama",
        "altair",
        "setuptools",
    ],
    entry_points={"console_scripts": ["fraggler=fraggler.cli:run"]},
)



from dataclasses import dataclass
import logging
from pathlib import Path

from ..applications.peak_area_multiplex import PeakAreaDeMultiplex
from ..utils.peak_finder import PeakFinder
from ..utils.fsa_file import FsaFile
from ..ladder_fitting.peak_ladder_assigner import PeakLadderAssigner
from ..ladder_fitting.fit_ladder_model import FitLadderModel


@dataclass
class FragglerPeak:
    fsa: FsaFile
    ladder_assigner: PeakLadderAssigner
    model: FitLadderModel
    peaks: PeakFinder


@dataclass
class FragglerArea:
    fsa: FsaFile
    ladder_assigner: PeakLadderAssigner
    model: FitLadderModel
    peaks: PeakFinder
    areas: PeakAreaDeMultiplex


def make_fraggler_peak(
    file: Path,
    ladder: str,
    min_height: int,
    min_ratio: float,
    trace_channel: str,
    peak_height: int,
    custom_peaks: str | None,
    size_standard_channel: str | None,
    distance_between_assays: int,
) -> FragglerPeak | str:
    file = Path(file)
    try:
        fsa = FsaFile(
            file,
            ladder,
            min_height=min_height,
            trace_channel=trace_channel,
            size_standard_channel=size_standard_channel,
        )
        ladder_assigner = PeakLadderAssigner(fsa)
        model = FitLadderModel(ladder_assigner)
        peaks = PeakFinder(
            model,
            min_ratio=min_ratio,
            peak_height=peak_height,
            custom_peaks=custom_peaks,
            distance_between_assays=distance_between_assays,
        )

        return FragglerPeak(
            fsa=fsa,
            ladder_assigner=ladder_assigner,
            model=model,
            peaks=peaks,
        )

    except Exception as e:
        logging.error(
            f"""Following file did not work: {file.stem}
        Reason: {e}
        """
        )

        return file.stem


def make_fraggler_area(
    file: Path,
    ladder: str,
    min_height: int,
    min_ratio: float,
    trace_channel: str,
    peak_height: int,
    custom_peaks: str | None,
    size_standard_channel: str | None,
    distance_between_assays: int,
    cutoff: int,
) -> FragglerArea | str:
    file = Path(file)
    try:
        fsa = FsaFile(
            file,
            ladder,
            min_height=min_height,
            trace_channel=trace_channel,
            size_standard_channel=size_standard_channel,
        )
        ladder_assigner = PeakLadderAssigner(fsa)
        model = FitLadderModel(ladder_assigner)
        peaks = PeakFinder(
            model,
            min_ratio=min_ratio,
            peak_height=peak_height,
            custom_peaks=custom_peaks,
            distance_between_assays=distance_between_assays,
        )
        areas = PeakAreaDeMultiplex(
            peaks,
            cutoff,
        )

        return FragglerArea(
            fsa=fsa,
            ladder_assigner=ladder_assigner,
            model=model,
            peaks=peaks,
            areas=areas,
        )

    except Exception as e:
        logging.error(
            f"""Following file did not work: {file.stem}
        Reason: {e}
        """
        )

        return file.stem



from fraggler import PeakAreaDeMultiplex

import matplotlib.pyplot as plt
import numpy as np


class PlotPeakArea:
    def __init__(self, peak_area: PeakAreaDeMultiplex):
        self.peak_area = peak_area

    def plot_areas(self, peak_finding_model: str, assay_number: int):

        # TODO: this functions get called again here...
        self.peak_area.fit_assay_peaks(peak_finding_model, assay_number, name="")

        fig_areas, axs = plt.subplots(
            1, len(self.peak_area.fit_df), sharey=True, figsize=(20, 10)
        )

        # if there is only one peak
        if len(self.peak_area.fit_df) == 1:
            axs.plot(
                self.peak_area.fit_df[0].basepairs, self.peak_area.fit_df[0].peaks, "o"
            )
            axs.plot(
                self.peak_area.fit_df[0].basepairs, self.peak_area.fit_df[0].fitted
            )
            axs.set_title(
                f"Peak 1 area: {self.peak_area.fit_params[0]['amplitude']: .1f}"
            )
            axs.grid()
        # if more than one peak
        else:
            for i, ax in enumerate(axs):
                ax.plot(
                    self.peak_area.fit_df[i].basepairs,
                    self.peak_area.fit_df[i].peaks,
                    "o",
                )
                ax.plot(
                    self.peak_area.fit_df[i].basepairs, self.peak_area.fit_df[i].fitted
                )
                ax.set_title(
                    f"Peak {i + 1} area: {self.peak_area.fit_params[i]['amplitude']: .1f}"
                )
                ax.grid()

        fig_areas.suptitle(f"Quotient: {self.peak_area.quotient: .2f}")
        fig_areas.legend(["Raw data", "Model"])
        fig_areas.supxlabel("basepairs")
        fig_areas.supylabel("intensity")
        plt.close()

        return fig_areas



from pathlib import Path
import logging
from datetime import datetime


def get_files(in_path: str) -> list[Path]:
    # If in_path is a directory, get a list of all .fsa files in it
    if Path(in_path).is_dir():
        files = [x for x in Path(in_path).iterdir() if x.suffix == ".fsa"]
    else:
        files = [Path(in_path)]
    return files


def setup_logging(outdir: str) -> None:
    """
    Set up the logging object and saves the log file to the same dir as the results files.
    """
    NOW = datetime.now().strftime("%Y-%m-%d_%H:%M")

    if not (outdir := Path(outdir)).exists():
        outdir.mkdir(parents=True)

    LOG_FILE = f"{outdir}/fraggler_{NOW}.log"

    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(levelname)s] \n%(message)s",
        datefmt="%Y-%m-%d %I:%M:%S",
        handlers=[logging.FileHandler(LOG_FILE), logging.StreamHandler()],
    )
